@@ -1,18 +1,15 @@
-use std::env;
-use std::fs;
 use std::io;
-use std::path::PathBuf;
 
 use anyhow::Result;
-use netrc_rs::Netrc;
 use reqwest::blocking::{Request, Response};
 use reqwest::header::{HeaderValue, AUTHORIZATION, WWW_AUTHENTICATE};
 use reqwest::StatusCode;
 
 use crate::cli::AuthType;
 use crate::middleware::{Context, Middleware};
+use crate::netrc;
 use crate::regex;
-use crate::utils::{clone_request, get_home_dir};
+use crate::utils::clone_request;
 
 #[derive(Debug, PartialEq, Eq)]
 pub enum Auth {
@@ -39,27 +36,22 @@ impl Auth {
         }
     }
 
-    pub fn from_netrc(netrc: &str, auth_type: AuthType, host: &str) -> Option<Auth> {
-        Netrc::parse_borrow(&netrc, false)
-            .ok()?
-            .machines
-            .into_iter()
-            .filter_map(|machine| match machine.name {
-                Some(name) if name == host => {
-                    let username = machine.login.unwrap_or_else(|| "".into());
-                    let password = machine.password;
-                    match auth_type {
-                        AuthType::basic => Some(Auth::Basic(username, password)),
-                        AuthType::digest => Some(Auth::Digest(
-                            username,
-                            password.unwrap_or_else(|| "".into()),
-                        )),
-                        AuthType::bearer => None,
-                    }
-                }
-                _ => None,
-            })
-            .last()
+    pub fn from_netrc(auth_type: AuthType, entry: netrc::Entry) -> Option<Auth> {
+        match auth_type {
+            AuthType::basic => Some(Auth::Basic(entry.login, Some(entry.password))),
+            AuthType::bearer => None,
+            AuthType::digest => Some(Auth::Digest(entry.login, entry.password)),
+        }
+    }
+
+    // Could be a method on AuthType instead, but it's helpful
+    // to put next to `Auth::from_netrc`
+    pub fn supports_netrc(auth_type: AuthType) -> bool {
+        match auth_type {
+            AuthType::basic => true,
+            AuthType::bearer => false,
+            AuthType::digest => true,
+        }
     }
 }
 
@@ -78,28 +70,6 @@ pub fn parse_auth(auth: &str, host: &str) -> io::Result<(String, Option<String>)
     }
 }
 
-pub fn read_netrc() -> Option<String> {
-    let netrc_path = match env::var_os("NETRC") {
-        Some(path) => {
-            let path = PathBuf::from(path);
-            if path.exists() {
-                Some(path)
-            } else {
-                None
-            }
-        }
-        None => {
-            let home_dir = get_home_dir()?;
-            [".netrc", "_netrc"]
-                .iter()
-                .map(|f| home_dir.join(f))
-                .find(|p| p.exists())
-        }
-    }?;
-
-    fs::read_to_string(netrc_path).ok()
-}
-
 pub struct DigestAuthMiddleware<'a> {
     username: &'a str,
     password: &'a str,
@@ -154,36 +124,4 @@ mod tests {
             assert_eq!(output, (user.as_str(), pass.as_deref()));
         }
     }
-
-    #[test]
-    fn netrc() {
-        let good_netrc = "machine example.com\nlogin user\npassword pass";
-        let malformed_netrc = "I'm a malformed netrc!";
-        let missing_login = "machine example.com\npassword pass";
-        let missing_pass = "machine example.com\nlogin user\n";
-
-        let expected = vec![
-            (
-                "example.com",
-                good_netrc,
-                Some(Auth::Basic("user".to_string(), Some("pass".to_string()))),
-            ),
-            ("example.org", good_netrc, None),
-            ("example.com", malformed_netrc, None),
-            (
-                "example.com",
-                missing_login,
-                Some(Auth::Basic("".to_string(), Some("pass".to_string()))),
-            ),
-            (
-                "example.com",
-                missing_pass,
-                Some(Auth::Basic("user".to_string(), None)),
-            ),
-        ];
-
-        for (machine, netrc, output) in expected {
-            assert_eq!(output, Auth::from_netrc(netrc, AuthType::basic, machine));
-        }
-    }
 }
