@@ -1,12 +1,12 @@
 use std::io::{stderr, stdout, Write};
 
 use anyhow::{anyhow, Result};
-use reqwest::Method;
+use os_display::Quotable;
+use reqwest::{tls, Method};
+use std::ffi::OsString;
 
-use crate::{
-    cli::{Cli, Verify},
-    request_items::{Body, RequestItem, FORM_CONTENT_TYPE, JSON_ACCEPT, JSON_CONTENT_TYPE},
-};
+use crate::cli::{AuthType, Cli, HttpVersion, Verify};
+use crate::request_items::{Body, RequestItem, FORM_CONTENT_TYPE, JSON_ACCEPT, JSON_CONTENT_TYPE};
 
 pub fn print_curl_translation(args: Cli) -> Result<()> {
     let cmd = translate(args)?;
@@ -23,7 +23,7 @@ pub fn print_curl_translation(args: Cli) -> Result<()> {
 
 pub struct Command {
     pub long: bool,
-    pub args: Vec<String>,
+    pub args: Vec<OsString>,
     pub env: Vec<(&'static str, String)>,
     pub warnings: Vec<String>,
 }
@@ -38,51 +38,42 @@ impl Command {
         }
     }
 
-    fn flag(&mut self, short: &'static str, long: &'static str) {
+    fn opt(&mut self, short: &'static str, long: &'static str) {
         if self.long {
-            self.args.push(long.to_string());
+            self.args.push(long.into());
         } else {
-            self.args.push(short.to_string());
+            self.args.push(short.into());
         }
     }
 
-    fn push(&mut self, arg: impl Into<String>) {
+    fn arg(&mut self, arg: impl Into<OsString>) {
         self.args.push(arg.into());
     }
 
     fn header(&mut self, name: &str, value: &str) {
-        self.flag("-H", "--header");
-        self.push(format!("{}: {}", name, value));
+        self.opt("-H", "--header");
+        self.arg(format!("{}: {}", name, value));
     }
 
     fn env(&mut self, var: &'static str, value: impl Into<String>) {
         self.env.push((var, value.into()));
     }
 
-    fn warn(&mut self, message: String) {
-        self.warnings.push(message);
+    fn warn(&mut self, message: impl Into<String>) {
+        self.warnings.push(message.into());
     }
 }
 
 impl std::fmt::Display for Command {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        let escape = if f.alternate() {
-            // If formatted with `{:#}`, use cmd.exe-style formatting
-            // This is currently not exposed
-            shell_escape::windows::escape
-        } else {
-            // By default, use Unix-style formatting regardless of platform
-            // This is also more suitable for Powershell
-            shell_escape::unix::escape
-        };
         for (key, value) in &self.env {
             // This is wrong for Windows, but there doesn't seem to be a
             // right way
-            write!(f, "{}={} ", key, escape(value.into()))?;
+            write!(f, "{}={} ", key, value.maybe_quote())?;
         }
         write!(f, "curl")?;
         for arg in &self.args {
-            write!(f, " {}", escape(arg.into()))?;
+            write!(f, " {}", arg.maybe_quote().external(true))?;
         }
         Ok(())
     }
@@ -94,12 +85,31 @@ pub fn translate(args: Cli) -> Result<Command> {
     let mut cmd = Command::new(args.curl_long);
 
     let ignored = &[
-        (args.offline, "--offline"),          // No equivalent
-        (args.body, "-b/--body"),             // Already the default
-        (args.print.is_some(), "-p/--print"), // No straightforward equivalent
-        (args.quiet, "-q/--quiet"),           // No equivalent, -s/--silent suppresses other stuff
-        (args.pretty.is_some(), "--pretty"),  // No equivalent
-        (args.style.is_some(), "-s/--style"), // No equivalent
+        // No equivalent
+        (args.offline, "--offline"),
+        // Already the default
+        (args.body, "-b/--body"),
+        // No straightforward equivalent
+        (args.print.is_some(), "-p/--print"),
+        // No equivalent, -s/--silent suppresses other stuff
+        (args.quiet, "-q/--quiet"),
+        // No equivalent
+        (args.pretty.is_some(), "--pretty"),
+        // No equivalent
+        (args.style.is_some(), "-s/--style"),
+        // No equivalent
+        (args.response_charset.is_some(), "--response-charset"),
+        // No equivalent
+        (args.response_mime.is_some(), "--response-mime"),
+        // Already the default
+        (args.all, "--all"),
+        // No (straightforward?) equivalent
+        (args.history_print.is_some(), "-P/--history-print"),
+        // Might be possible to emulate with --cookie-jar but tricky
+        (args.session.is_some(), "--session"),
+        // Already the default (usually, depends on compile time options)
+        // Unclear if you can even change this at runtime
+        (args.native_tls, "--native-tls"),
     ];
 
     for (present, flag) in ignored {
@@ -116,65 +126,87 @@ pub fn translate(args: Cli) -> Result<Command> {
     // Output options
     if args.verbose {
         // Far from an exact match, but it does print the request headers
-        cmd.flag("-v", "--verbose");
+        cmd.opt("-v", "--verbose");
     }
     if args.stream {
         // curl sorta streams by default, but its buffer stops it from
         // showing up right away
-        cmd.flag("-N", "--no-buffer");
+        cmd.opt("-N", "--no-buffer");
     }
-    // Since --fail is bit disruptive than HTTPie's --check-status flag, we will not enable
+    // Since --fail is more disruptive than HTTPie's --check-status flag, we will not enable
     // it unless the user explicitely sets the latter flag
-    if args.check_status.unwrap_or(false) {
+    if args.check_status == Some(true) {
         // Suppresses output on failure, unlike us
-        cmd.flag("-f", "--fail");
+        cmd.opt("-f", "--fail");
     }
 
     // HTTP options
     if args.follow {
-        cmd.flag("-L", "--location");
+        cmd.opt("-L", "--location");
     }
     if let Some(num) = args.max_redirects {
-        cmd.push("--max-redirects");
-        cmd.push(num.to_string());
+        cmd.arg("--max-redirs");
+        cmd.arg(num.to_string());
     }
     if let Some(filename) = args.output {
         let filename = filename.to_str().ok_or_else(|| anyhow!("Invalid UTF-8"))?;
-        cmd.flag("-o", "--output");
-        cmd.push(filename);
+        cmd.opt("-o", "--output");
+        cmd.arg(filename);
     } else if args.download {
-        cmd.flag("-O", "--remote-name");
+        cmd.opt("-O", "--remote-name");
     }
     if args.resume {
-        cmd.flag("-C", "--continue-at");
-        cmd.push("-"); // Tell curl to guess, like we do
+        cmd.opt("-C", "--continue-at");
+        cmd.arg("-"); // Tell curl to guess, like we do
     }
     match args.verify.unwrap_or(Verify::Yes) {
         Verify::CustomCaBundle(filename) => {
-            cmd.push("--cacert");
-            // TODO: maybe filename should be as bytes?
-            // (does the way we have structopt set up even accept non-unicode?)
-            cmd.push(filename.to_string_lossy());
+            cmd.arg("--cacert");
+            cmd.arg(filename);
         }
         Verify::No => {
-            cmd.flag("-k", "--insecure");
+            cmd.opt("-k", "--insecure");
         }
         Verify::Yes => {}
     }
     if let Some(cert) = args.cert {
-        cmd.flag("-E", "--cert");
-        // TODO: as bytes?
-        cmd.push(cert.to_string_lossy());
+        cmd.opt("-E", "--cert");
+        cmd.arg(cert);
     }
     if let Some(keyfile) = args.cert_key {
-        cmd.push("--key");
-        cmd.push(keyfile.to_string_lossy());
+        cmd.arg("--key");
+        cmd.arg(keyfile);
+    }
+    if let Some(Some(tls_version)) = args.ssl {
+        match tls_version {
+            tls::Version::TLS_1_0 => {
+                cmd.arg("--tlsv1.0");
+                cmd.arg("--tls-max");
+                cmd.arg("1.0");
+            }
+            tls::Version::TLS_1_1 => {
+                cmd.arg("--tlsv1.1");
+                cmd.arg("--tls-max");
+                cmd.arg("1.1");
+            }
+            tls::Version::TLS_1_2 => {
+                cmd.arg("--tlsv1.2");
+                cmd.arg("--tls-max");
+                cmd.arg("1.2");
+            }
+            tls::Version::TLS_1_3 => {
+                cmd.arg("--tlsv1.3");
+                cmd.arg("--tls-max");
+                cmd.arg("1.3");
+            }
+            _ => unreachable!(),
+        }
     }
     for proxy in args.proxy {
         match proxy {
             crate::cli::Proxy::All(proxy) => {
-                cmd.flag("-x", "--proxy");
-                cmd.push(proxy);
+                cmd.opt("-x", "--proxy");
+                cmd.arg(String::from(proxy));
             }
             crate::cli::Proxy::Http(proxy) => {
                 // These don't seem to have corresponding flags
@@ -185,14 +217,25 @@ pub fn translate(args: Cli) -> Result<Command> {
             }
         }
     }
+    if let Some(timeout) = args.timeout.and_then(|t| t.as_duration()) {
+        cmd.arg("--max-time");
+        cmd.arg(timeout.as_secs_f64().to_string());
+    }
+    if let Some(http_version) = args.http_version {
+        match http_version {
+            HttpVersion::Http10 => cmd.arg("--http1.0"),
+            HttpVersion::Http11 => cmd.arg("--http1.1"),
+            HttpVersion::Http2 => cmd.arg("--http2"),
+        }
+    }
 
     if args.method == Some(Method::HEAD) {
-        cmd.flag("-I", "--head");
+        cmd.opt("-I", "--head");
     } else if args.method == Some(Method::OPTIONS) {
         // If you're sending an OPTIONS you almost certainly want to see the headers
-        cmd.flag("-i", "--include");
-        cmd.flag("-X", "--request");
-        cmd.push("OPTIONS");
+        cmd.opt("-i", "--include");
+        cmd.opt("-X", "--request");
+        cmd.arg("OPTIONS");
     } else if args.headers {
         // The best option for printing just headers seems to be to use -I
         // but with an explicit method as an override.
@@ -204,9 +247,9 @@ pub fn translate(args: Cli) -> Result<Command> {
             // unwrap_or_else causes borrowing issues
             None => args.request_items.pick_method(),
         };
-        cmd.flag("-I", "--head");
-        cmd.flag("-X", "--request");
-        cmd.push(method.to_string());
+        cmd.opt("-I", "--head");
+        cmd.opt("-X", "--request");
+        cmd.arg(method.to_string());
         if method != Method::GET {
             cmd.warn(
                 "-I/--head is incompatible with sending data. Consider omitting -h/--headers."
@@ -214,35 +257,50 @@ pub fn translate(args: Cli) -> Result<Command> {
             );
         }
     } else if let Some(method) = args.method {
-        cmd.flag("-X", "--request");
-        cmd.push(method.to_string());
+        cmd.opt("-X", "--request");
+        cmd.arg(method.to_string());
     }
     // We assume that curl's automatic detection of when to do a POST matches
     // ours so we can ignore the None case
 
-    cmd.push(args.url.to_string());
+    cmd.arg(args.url.to_string());
 
     // Payload
     for (header, value) in headers.iter() {
-        cmd.flag("-H", "--header");
+        cmd.opt("-H", "--header");
         if value.is_empty() {
-            cmd.push(format!("{};", header));
+            cmd.arg(format!("{};", header));
         } else {
-            cmd.push(format!("{}: {}", header, value.to_str()?));
+            cmd.arg(format!("{}: {}", header, value.to_str()?));
         }
     }
     for header in headers_to_unset {
-        cmd.flag("-H", "--header");
-        cmd.push(format!("{}:", header));
+        cmd.opt("-H", "--header");
+        cmd.arg(format!("{}:", header));
+    }
+    if args.ignore_netrc {
+        // Already the default, so a bit questionable
+        cmd.arg("--no-netrc");
     }
     if let Some(auth) = args.auth {
-        // curl implements this flag the same way, including password prompt
-        cmd.flag("-u", "--user");
-        cmd.push(auth);
-    }
-    if let Some(token) = args.bearer {
-        cmd.push("--oauth2-bearer");
-        cmd.push(token);
+        match args.auth_type.unwrap_or_default() {
+            AuthType::basic => {
+                cmd.arg("--basic");
+                // curl implements this flag the same way, including password prompt
+                cmd.opt("-u", "--user");
+                cmd.arg(auth);
+            }
+            AuthType::digest => {
+                cmd.arg("--digest");
+                // curl implements this flag the same way, including password prompt
+                cmd.opt("-u", "--user");
+                cmd.arg(auth);
+            }
+            AuthType::bearer => {
+                cmd.arg("--oauth2-bearer");
+                cmd.arg(auth);
+            }
+        }
     }
 
     if args.request_items.is_multipart() {
@@ -254,12 +312,12 @@ pub fn translate(args: Cli) -> Result<Command> {
                     return Err(anyhow!("JSON values are not supported in multipart fields"));
                 }
                 RequestItem::DataField(key, value) => {
-                    cmd.flag("-F", "--form");
-                    cmd.push(format!("{}={}", key, value));
+                    cmd.opt("-F", "--form");
+                    cmd.arg(format!("{}={}", key, value));
                 }
                 RequestItem::DataFieldFromFile(key, value) => {
-                    cmd.flag("-F", "--form");
-                    cmd.push(format!("{}=<{}", key, value));
+                    cmd.opt("-F", "--form");
+                    cmd.arg(format!("{}=<{}", key, value));
                 }
                 RequestItem::FormFile {
                     key,
@@ -267,7 +325,7 @@ pub fn translate(args: Cli) -> Result<Command> {
                     file_type,
                     file_name_header,
                 } => {
-                    cmd.flag("-F", "--form");
+                    cmd.opt("-F", "--form");
                     let mut val = format!("{}=@{}", key, file_name);
                     if let Some(file_type) = file_type {
                         val.push_str(";type=");
@@ -277,7 +335,7 @@ pub fn translate(args: Cli) -> Result<Command> {
                         val.push_str(";filename=");
                         val.push_str(&file_name_header);
                     }
-                    cmd.push(val);
+                    cmd.arg(val);
                 }
                 RequestItem::HttpHeader(..) => {}
                 RequestItem::HttpHeaderToUnset(..) => {}
@@ -295,12 +353,12 @@ pub fn translate(args: Cli) -> Result<Command> {
                     // More faithful than -F, but doesn't have a short version
                     // New in curl 7.18.0 (January 28 2008), *probably* old enough
                     // Otherwise passing --multipart helps
-                    cmd.push("--data-urlencode");
+                    cmd.arg("--data-urlencode");
                     // Encoding this is tricky: --data-urlencode expects name
                     // to be encoded but not value and doesn't take strings
                     let mut encoded = serde_urlencoded::to_string(&[(key, "")])?;
                     encoded.push_str(&value);
-                    cmd.push(encoded);
+                    cmd.arg(encoded);
                 }
             }
             Body::Json(map) if !map.is_empty() => {
@@ -308,8 +366,8 @@ pub fn translate(args: Cli) -> Result<Command> {
                 cmd.header("accept", JSON_ACCEPT);
 
                 let json_string = serde_json::Value::from(map).to_string();
-                cmd.flag("-d", "--data");
-                cmd.push(json_string);
+                cmd.opt("-d", "--data");
+                cmd.arg(json_string);
             }
             Body::Json(..) if args.json => {
                 cmd.header("content-type", JSON_CONTENT_TYPE);
@@ -328,8 +386,10 @@ pub fn translate(args: Cli) -> Result<Command> {
                 } else {
                     cmd.header("content-type", JSON_CONTENT_TYPE);
                 }
-                cmd.push("--data-binary");
-                cmd.push(format!("@{}", file_name.to_string_lossy()));
+                cmd.arg("--data-binary");
+                let mut arg = OsString::from("@");
+                arg.push(file_name);
+                cmd.arg(arg);
             }
         }
     }
@@ -344,118 +404,97 @@ mod tests {
     #[test]
     fn examples() {
         let expected = vec![
-            ("xh httpbin.org/get", "curl 'http://httpbin.org/get'", "curl http://httpbin.org/get"),
+            ("xh httpbin.org/get", "curl http://httpbin.org/get"),
             (
                 "xh httpbin.org/post x=3",
-                r#"curl 'http://httpbin.org/post' -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5' -d '{"x":"3"}'"#,
-                r#"curl http://httpbin.org/post -H "content-type: application/json" -H "accept: application/json, */*;q=0.5" -d "{\"x\":\"3\"}""#,
+                #[cfg(not(windows))]
+                r#"curl http://httpbin.org/post -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5' -d '{"x":"3"}'"#,
+                #[cfg(windows)]
+                r#"curl http://httpbin.org/post -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5' -d '{\"x\":\"3\"}'"#,
             ),
             (
                 "xh --form httpbin.org/post x\\=y=z=w",
-                "curl 'http://httpbin.org/post' --data-urlencode 'x%3Dy=z=w'",
-                "curl http://httpbin.org/post --data-urlencode x%3Dy=z=w",
+                "curl http://httpbin.org/post --data-urlencode 'x%3Dy=z=w'",
             ),
             (
                 "xh put httpbin.org/put",
-                "curl -X PUT 'http://httpbin.org/put'",
                 "curl -X PUT http://httpbin.org/put",
             ),
             (
                 "xh --https httpbin.org/get x==3",
                 "curl 'https://httpbin.org/get?x=3'",
-                "curl https://httpbin.org/get?x=3",
             ),
             (
                 "xhs httpbin.org/get x==3",
                 "curl 'https://httpbin.org/get?x=3'",
-                "curl https://httpbin.org/get?x=3",
             ),
             (
                 "xh -h httpbin.org/get",
-                "curl -I -X GET 'http://httpbin.org/get'",
                 "curl -I -X GET http://httpbin.org/get",
             ),
             (
                 "xh options httpbin.org/get",
-                "curl -i -X OPTIONS 'http://httpbin.org/get'",
                 "curl -i -X OPTIONS http://httpbin.org/get",
             ),
             (
                 "xh --proxy http:localhost:1080 httpbin.org/get",
-                "http_proxy='localhost:1080' curl 'http://httpbin.org/get'",
                 "http_proxy=localhost:1080 curl http://httpbin.org/get",
             ),
             (
                 "xh --proxy all:localhost:1080 httpbin.org/get",
-                "curl -x 'localhost:1080' 'http://httpbin.org/get'",
                 "curl -x localhost:1080 http://httpbin.org/get",
             ),
             (
                 "xh httpbin.org/post x:=[3]",
-                r#"curl 'http://httpbin.org/post' -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5' -d '{"x":[3]}'"#,
-                r#"curl http://httpbin.org/post -H "content-type: application/json" -H "accept: application/json, */*;q=0.5" -d "{\"x\":[3]}""#,
+                #[cfg(not(windows))]
+                r#"curl http://httpbin.org/post -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5' -d '{"x":[3]}'"#,
+                #[cfg(windows)]
+                r#"curl http://httpbin.org/post -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5' -d '{\"x\":[3]}'"#,
             ),
             (
                 "xh --json httpbin.org/post",
-                "curl 'http://httpbin.org/post' -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5'",
-                r#"curl http://httpbin.org/post -H "content-type: application/json" -H "accept: application/json, */*;q=0.5""#,
+                "curl http://httpbin.org/post -H 'content-type: application/json' -H 'accept: application/json, */*;q=0.5'",
             ),
             (
                 "xh --form httpbin.org/post x@/dev/null",
-                "curl 'http://httpbin.org/post' -F 'x=@/dev/null'",
-                "curl http://httpbin.org/post -F x=@/dev/null",
+                "curl http://httpbin.org/post -F 'x=@/dev/null'",
             ),
             (
                 "xh --form httpbin.org/post",
-                "curl 'http://httpbin.org/post' -H 'content-type: application/x-www-form-urlencoded'",
-                r#"curl http://httpbin.org/post -H "content-type: application/x-www-form-urlencoded""#,
+                "curl http://httpbin.org/post -H 'content-type: application/x-www-form-urlencoded'",
             ),
             (
                 "xh --bearer foobar post httpbin.org/post",
-                "curl -X POST 'http://httpbin.org/post' --oauth2-bearer foobar",
                 "curl -X POST http://httpbin.org/post --oauth2-bearer foobar",
             ),
             (
                 "xh httpbin.org/get foo:Bar baz; user-agent:",
-                "curl 'http://httpbin.org/get' -H 'foo: Bar' -H 'baz;' -H 'user-agent:'",
-                r#"curl http://httpbin.org/get -H "foo: Bar" -H baz; -H user-agent:"#,
+                "curl http://httpbin.org/get -H 'foo: Bar' -H 'baz;' -H user-agent:",
             ),
             (
                 "xh -d httpbin.org/get",
-                "curl -f -L -O 'http://httpbin.org/get'",
                 "curl -f -L -O http://httpbin.org/get",
             ),
             (
                 "xh -d -o foobar --continue httpbin.org/get",
-                "curl -f -L -o foobar -C - 'http://httpbin.org/get'",
                 "curl -f -L -o foobar -C - http://httpbin.org/get",
             ),
             (
                 "xh --curl-long -d -o foobar --continue httpbin.org/get",
-                "curl --fail --location --output foobar --continue-at - 'http://httpbin.org/get'",
                 "curl --fail --location --output foobar --continue-at - http://httpbin.org/get",
             ),
             (
                 "xh httpbin.org/post @foo.txt",
-                "curl 'http://httpbin.org/post' -H 'content-type: text/plain' --data-binary '@foo.txt'",
-                r#"curl http://httpbin.org/post -H "content-type: text/plain" --data-binary @foo.txt"#,
+                #[cfg(not(windows))]
+                "curl http://httpbin.org/post -H 'content-type: text/plain' --data-binary @foo.txt",
+                #[cfg(windows)]
+                "curl http://httpbin.org/post -H 'content-type: text/plain' --data-binary '@foo.txt'",
             ),
         ];
-        for (input, output_unix, output_windows) in expected {
+        for (input, output) in expected {
             let cli = Cli::from_iter_safe(input.split_whitespace()).unwrap();
             let cmd = translate(cli).unwrap();
-            assert_eq!(
-                cmd.to_string(),
-                output_unix,
-                "Wrong Unix output for {:?}",
-                input
-            );
-            assert_eq!(
-                format!("{:#}", cmd),
-                output_windows,
-                "Wrong Windows output for {:?}",
-                input
-            );
+            assert_eq!(cmd.to_string(), output, "Wrong output for {:?}", input);
         }
     }
 }
