@@ -1,6 +1,6 @@
 use std::env::var_os;
 use std::io::{self, Write};
-use std::path::PathBuf;
+use std::path::{Path, PathBuf};
 
 use anyhow::Result;
 use reqwest::blocking::Request;
@@ -59,6 +59,22 @@ pub fn get_home_dir() -> Option<PathBuf> {
     dirs::home_dir()
 }
 
+/// Perform simple tilde expansion if `dirs::home_dir()` is `Some(path)`.
+///
+/// Note that prefixed tilde e.g `~foo` is ignored.
+///
+/// See https://www.gnu.org/software/bash/manual/html_node/Tilde-Expansion.html
+pub fn expand_tilde(path: impl AsRef<Path>) -> PathBuf {
+    if let Ok(path) = path.as_ref().strip_prefix("~") {
+        let mut expanded_path = PathBuf::new();
+        expanded_path.push(get_home_dir().unwrap_or_else(|| "~".into()));
+        expanded_path.push(path);
+        expanded_path
+    } else {
+        path.as_ref().into()
+    }
+}
+
 // https://stackoverflow.com/a/45145246/5915221
 #[macro_export]
 macro_rules! vec_of_strings {
@@ -94,7 +110,6 @@ pub const BUFFER_SIZE: usize = 64 * 1024;
 /// served locally by `python3 -m http.server`. Results may vary.
 pub fn copy_largebuf(reader: &mut impl io::Read, writer: &mut impl Write) -> io::Result<()> {
     let mut buf = vec![0; BUFFER_SIZE];
-    let mut buf = buf.as_mut_slice();
     loop {
         match reader.read(&mut buf) {
             Ok(0) => return Ok(()),
